"""Cython-coded accelerators for the PyOpenGL wrapper

This package contains Cython accelerator modules which
attempt to speed up certain aspects of the PyOpenGL 3.x
wrapper mechanism.  The source code is part of the
PyOpenGL package and is built via the setupaccel.py
script in the top level of the PyOpenGL source package.
"""

__version__ = "3.1.9a2"
__version_tuple__ = (3, 1, 9)
